/*******************************************************************************
 * mod_authn_citadel.c: Citadel authentication module for apache
 *
 * Project Contact Information:
 *   Stuart Cianos
 *   Email: scianos@alphavida.com
 *
 *******************************************************************************
 * (C) Copyright 2008 by Stuart Cianos, d/b/a AlphaVida. All Rights Reserved.
 *
 *
 * Stuart Cianos licenses this file to You under the GNU General Public License,
 * Version 3.0 (the "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.gnu.org/licenses/gpl.txt
 *
 * or in the COPYING file included in the original archive.
 *
 * Disclaimer of Warranty.
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
 * HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
 * OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
 * IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
 * ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
 *
 * Limitation of Liability.
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
 * THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
 * GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
 * USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
 * DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
 * PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
 * EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGES.
 *******************************************************************************
 * DESCRIPTION:
 *
 * Apache authentication module for Citadel Groupware Server users. Allows users
 * to be seamlessly authenticated against the Citadel user database. Uses the
 * native Citadel IPC protocol; features SSL/TLS support for secure
 * communications to the Citadel server.
 *******************************************************************************
 * DEPENDENCIES:
 *
 * The following libraries and includes must be available to compile the
 * module. Primary dependencies include the Apache Portal Runtime (both apr and
 * apr_util) as well as the OpenSSL libraries.
 *  - Apache Server Development Files
 *  - Apache Portable Runtime - libraries and development files: apr, apr_util
 *  - OpenSSL libraries and development files
 *******************************************************************************
 * USAGE:
 *
 * The module must be loaded before use. Example:
 *  LoadModule authn_citadel_module /usr/local/lib/libmod_authn_citadel.so
 *
 * The path to the module on your system may vary, and should be adjusted
 * accordingly.
 *
 * The following configuration directives are used to configure Citadel
 * authentication in a valid Apache configuration block (i.e. location or
 * directory):
 *
 * CitadelServer <host>: The Citadel server to contact for authentication.
 *                       Can be either a fully qualified hostname or an IP
 *                       address.
 *
 * CitadelPort <port>:   The port that the Citadel server should be contacted
 *                       on. This is port 504 in most installations.
 *
 * CitadelTimeout <sec>: The number of seconds to time out communication.
 *                       This currently defaults to 2. Not all communication
 *                       routines are using this in current versions, but it
 *                       will be used in future revisions.
 *
 * CitadelUseSSL <y/n>:  Determines if the connection to Citadel will be
 *                       encrypted with SSL. The value passed in <y/n> can
 *                       be either "Yes" or "No". This should ALWAYS be set
 *                       to Yes if the Citadel server is hosted on a remote
 *                       server, or if the communication between the web server
 *                       and Citadel are traveling across an insecure network
 *                       which could be sniffed.
 *
 * CitadelMinLevel <l>:  The minimum access level required to be granted
 *                       access. The value <l> can be any single-digit value
 *                       between 0 and 6.
 *
 * As this is a basic authentication module, The AuthType should be set to
 * "basic" within the configuration block. The AuthBasicProvider is "citadel".
 *
 * Example configuration:
 *
 * <Directory /var/www/>
 *    AuthType basic
 *    AuthBasicProvider citadel
 *    AuthName Citadel Server
 *    Require valid-user
 *
 *    CitadelServer 127.0.0.1
 *    CitadelPort 504
 *    CitadelTimeout 2
 *    CitadelUseSSL Yes
 *    CitadelMinLevel 1
 *
 *    Options Indexes FollowSymLinks MultiViews
 *    AllowOverride None
 *    Order allow,deny
 *    allow from all
 * </Directory>
 *
 *******************************************************************************
 * REVISION HISTORY:
 *
 * 2008-03-05: Stuart Cianos <scianos@alphavida.com>
 *  - Initial ALPHA release, for testing only: Version 0.01 ALPHA
 *  - Do not use in a production environment. Contains known bugs.
 * 2008-08-31: Stuart Cianos <scianos@alphavida.com>
 *  - Initial BETA release, for testing only: Version 0.02 BETA
 *  - Completed SSL/TLS protocol support
 *  - Improved data read function; added protocol inteligence to ensure complete
 *    reads on multiline or large buffered reads.
 *  - Gracefully terminate the session when possible, instead of just dropping
 *    the connection
 *  - Cleaned up logging
 * 2008-09-05: Stuart Cianos <scianos@alphavida.com>
 *  - Cleaned up comment formatting
 *  - Added CitadelMinLevel functionality
 *  - Added routines to tokenize return parameters
 *  - Release 0.99-BETA1
 *
 ******************************************************************************/

/* Standard C headers */
#include <stdarg.h>
#include <time.h>

/* Include OPENSSL */
#ifndef HAVE_LIBSSL
#error OpenSSL is required for mod_auth_citadel
#else
#include "openssl/bio.h"
#include "openssl/ssl.h"
#include "openssl/err.h"
#endif

/* Include Apache Headers */
#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_log.h"
#include "http_request.h"
#include "mod_auth.h"

/* Are we DEBUGGING? */
#ifdef MODCITDEBUG
#define AUTHN_CITADEL_DEBUGMODE
#endif

/* Citadel Return Codes (primary) */
#define AUTHN_CITADEL_STAT_LISTING_FOLLOWS		100 /* List Follows   */
#define AUTHN_CITADEL_STAT_CIT_OK			200 /* OK             */
#define AUTHN_CITADEL_STAT_MORE_DATA			300 /* More Data      */
#define AUTHN_CITADEL_STAT_SEND_LISTING			400 /* Send Data      */
#define AUTHN_CITADEL_STAT_ERROR			500 /* Error          */
#define AUTHN_CITADEL_STAT_BINARY_FOLLOWS		600 /* More Bin       */
#define AUTHN_CITADEL_STAT_SEND_BINARY			700 /* Send Data      */
#define AUTHN_CITADEL_STAT_START_CHAT_MODE		800 /* Start Chat Mode*/

/* Protocol constants */
#define AUTHN_CITADEL_NETTIMEOUT			2	/* TODO: Conf */
#define AUTHN_CITADEL_LINEDELIM				"\n"    /* End of line*/
#define AUTHN_CITADEL_PARMDELIM                         "|"     /* Parameter  */

/* Version and indentification constants */
#define AUTHN_CITADEL_IDEN_DEVELOPER                    "31"
#define AUTHN_CITADEL_IDEN_CLIENTID                     "0001"
#define AUTHN_CITADEL_IDEN_VERSION                      "0.99-BETA1"
#define AUTHN_CITADEL_IDEN_DESCRIPTION                  "mod_authn_citadel"

/* Configuration structure */
typedef struct {
    char *citadelserver;    /* Citadel hostname */
    char *port;             /* Port that Citadel listens on */
    char *timeout;          /* Timeout - TODO: Use config, not constant */
    char *usessl;           /* Use SSL - "yes" or "no" */
    char *minlevel;         /* Minimum level - number between 0-6 */
} authn_citadel_config_rec;

/*
 * Read and configure the module
 */
static void *create_authn_citadel_dir_config(apr_pool_t *p, char *d) {
    authn_citadel_config_rec *conf = apr_palloc(p, sizeof(*conf));
    
    /* Set Defaults */
    conf->citadelserver = NULL;     /* No default server */
    conf->usessl = "yes";           /* Use SSL by default */
    conf->timeout = "2";            /* Default timeout is 2 seconds */
    conf->port = "504";             /* Default port number is 504 */
    conf->minlevel = "1";           /* Default minimum level is 1 */
    return conf;
}

/*
 * Set the minimum access level for successful authentication
 * in the config struct
 */
static const char *set_authn_citadel_set_citadelminlevel(cmd_parms *cmd,
        void *offset, const char *f, const char *t) {
    char fres[2];
    char x;
    
    fres[0] = 0;
    fres[1] = 0;
    
    /* Default minimum access level is 1 */
    x = 1;
    
    /* Convert to small int from character string */
    x = atoi(f);
    
    /* No access level greater than 6 is allowed */
    if (x > 6) {
        x = 6;
    }
    
    /* No access level less than 0 is allowed */
    if (x < 0) {
        x = 0;
    }
    
    /* Place single digit access level back into the configuration structure */
    snprintf(fres, 1, "%i*", x, 1);
    
    return ap_set_string_slot(cmd, offset, f);
}

/*
 * Set the citadel server in the config struct
 */
static const char *set_authn_citadel_set_citadelserver(cmd_parms *cmd,
        void *offset, const char *f, const char *t) {
    return ap_set_string_slot(cmd, offset, f);
}

/*
 * Set the citadel port in the config struct
 */
static const char *set_authn_citadel_set_citadelport(cmd_parms *cmd,
        void *offset, const char *f, const char *t) {
    return ap_set_string_slot(cmd, offset, f);;
}

/*
 * Set the citadel comm timeout (not implemented in functional code - TODO)
 */
static const char *set_authn_citadel_set_citadeltimeout(cmd_parms *cmd,
        void *offset, const char *f, const char *t) {
    return ap_set_string_slot(cmd, offset, f);
}

/*
 * Set the SSL status in the config struct
 */
static const char *set_authn_citadel_set_citadelusessl(cmd_parms *cmd,
        void *offset, const char *f, const char *t) {
    char *fres;
    
    /* Validate user input from config file */
    if (strcasecmp(f, "yes") == 0) {
        fres = "yes";
    } else {
        fres = "no";
    }
    
    return ap_set_string_slot(cmd, offset, fres);
}

/*
 * Macro to configure the module from the specified configuration
 * section by offset. Uses the Apache configuration macros.
 */
static const command_rec authn_citadel_cmds[] = {
    AP_INIT_TAKE1("CitadelServer", set_authn_citadel_set_citadelserver,
            (void *)APR_OFFSETOF(authn_citadel_config_rec, citadelserver),
            OR_AUTHCFG, "IP address or hostname of the Citadel server"),
    AP_INIT_TAKE1("CitadelPort", set_authn_citadel_set_citadelport,
            (void *)APR_OFFSETOF(authn_citadel_config_rec, port),
            OR_AUTHCFG, "Port of the Citadel server"),
    AP_INIT_TAKE1("CitadelTimeout", set_authn_citadel_set_citadeltimeout,
            (void *)APR_OFFSETOF(authn_citadel_config_rec, timeout),
            OR_AUTHCFG, "Timeout for communications with citadel"),
    AP_INIT_TAKE1("CitadelUseSSL", set_authn_citadel_set_citadelusessl,
            (void *)APR_OFFSETOF(authn_citadel_config_rec, usessl),
            OR_AUTHCFG,
            "Use STARTTLS for authentication. Turning this off is INSECURE!"),
    AP_INIT_TAKE1("CitadelMinLevel", set_authn_citadel_set_citadelminlevel,
            (void *)APR_OFFSETOF(authn_citadel_config_rec, minlevel),
            OR_AUTHCFG,
            "Minimum Citadel access level for access to be granted"),
    {NULL}
};

/*
 * Module structure declaration
 */
module AP_MODULE_DECLARE_DATA authn_citadel_module;

/*
 * Parse the Citadel protocol return code, from buffered results
 */
static int authn_citadel_parse_return_code(char *result) {
    char *tok;
    char code[4];
    int ret;
    char *ts;
    char *ts_state;
    
    /* Allocate working memory */
    ts = (void *) calloc(strlen(result) + 1, sizeof(char));
    memcpy(ts, result, strlen(result));
    
    /* Get the first line in the buffer */
    tok = strtok_r(ts, AUTHN_CITADEL_LINEDELIM, &ts_state);
    
    /* If we are not NULL, check last 3 characters in the first line */
    if (tok != NULL) {
        /* Check for >=3 characters, and validate that result is
         * numeric. */
        if (strlen(tok) >= 3 &&
                isdigit(tok[0]) && isdigit(tok[1]) && isdigit(tok[2])) {
            /* Copy numeric result into memory location holding
             * the code */
            memset(code, 0, 4);
            memcpy(code, tok, 3);
            
            /* Convert code into int */
            ret = atoi(code);
            
            /* Free the buffer allocated earlier */
            free(ts);
            
            /* Return the Citadel result code */
            return ret;
        }
    }
    
    /* If we get to this code, no result code was available or there
     * was some other problem. Free the previously allocated memory
     * and return -1. */
    free(ts);
    
    return -1;
}

/*
 * Parse the Citadel protocol return line, from buffered results, and
 * return the specified parameter into result.
 */
static int authn_citadel_parse_return_parm(char *outbuf, int param,
        char *result, int size) {
    char *tok;
    char *ptok;
    int ret;
    char *ts;
    char *ts_state;
    char *pts_state;
    int pnum;
    
    /* Allocate working memory */
    ts = (void *) calloc(strlen(result) + 1, sizeof(char));
    memcpy(ts, result, strlen(result));
    
    /* Clear output buffer */
    memset(outbuf, 0, size);
    
    /* Set default return value */
    ret = -1;
    
    
    
    /* Get the first line in the buffer */
    tok = strtok_r(ts, AUTHN_CITADEL_LINEDELIM, &ts_state);
    
    /* If we are not NULL, check first 3 characters in the first line */
    if (tok != NULL) {
        /* Check for >=3 characters, and validate that result is
         * numeric. */
        if (strlen(tok) >= 3 &&
                isdigit(tok[0]) && isdigit(tok[1]) && isdigit(tok[2])) {
            
            /* Get first parameter */
            ptok = strtok_r(ts, AUTHN_CITADEL_PARMDELIM, &pts_state);
            
            /* Initialize parameter count */
            pnum = 0;
            
            /* Loop through parameters */
            while (ptok != NULL) {
                pnum++;  /* Increment parameter count */
                
                if (pnum == param) {
                    /* Copy parameter text into output buffer */
                    strncpy(outbuf, ptok, size-1);
                    
                    /* Return the size of the returned parameters */
                    ret = strlen(outbuf);
                    break;
                }
                
                /* Get next parameter */
                ptok = strtok_r(NULL, AUTHN_CITADEL_PARMDELIM, &pts_state);
            }
            
            /* Free the buffer allocated earlier */
            free(ts);
            
            /* Return the -1 if error, or length of returned string */
            return ret;
        }
    }
    
    /* If we get to this code, no result code was available or there
     * was some other problem. Free the previously allocated memory
     * and return -1. */
    free(ts);
    
    return -1;
}


/*
 * Parse results buffer, checking for listing complete/data complete code
 * in Citadel protocol.
 */
static int authn_citadel_parse_listing_over(char *result) {
    char *tok;
    char code[4];
    char *ts;
    char *ts_state;
    char *tloc;
    
    /* Allocate memory for temporary buffer */
    ts = (void *) calloc(strlen(result) + 1, sizeof(char));
    
    /* Copy the results into the temporary buffer */
    memcpy(ts, result, strlen(result));
    
    /* Tokenize temporary buffer by line. Set pointer to the first
     * line. */
    tok = strtok_r(ts, AUTHN_CITADEL_LINEDELIM, &ts_state);
    
    /* Loop while data in buffer remains to be parsed */
    while (tok != NULL) {
        /* Check to ensure that there are at least 3 characters in the
         * line */
        if (strlen(tok) >= 3) {
            /* Copy the first three characters into temporary
             * code location */
            memset(code, 0, 4);
            memcpy(code, tok, 3);
            
            /* Compare the code. If it is "000", the data
             * transmission is complete. */
            if (strcmp(code, "000") == 0) {
                /* Free the previously allocated memory */
                free(ts);
                return TRUE;
            }
        }
        
        /* Get next line */
        tok = strtok_r(NULL, AUTHN_CITADEL_LINEDELIM, &ts_state);
        
    }
    
    /* If code at this location is executing, the end of data has not
     * been reached. Free the previously allocated memory, and return
     * FALSE to the calling procedure. */
    free(ts);
    
    return FALSE;
}

/*
 * Return a line from a citadel protocol list
 */
static int authn_citadel_parse_list_line(char *outbuf, int line, char *result,
        int size) {
    char *tok;
    char code[4];
    char *ts;
    char *ts_state;
    char *tloc;
    
    /* Allocate memory */
    ts = (void *) calloc(strlen(result) + 1, sizeof(char));
    memcpy(ts, result, strlen(result));
    
    /* Tokenize the buffer */
    tok = strtok_r(ts, AUTHN_CITADEL_LINEDELIM, &ts_state);
    
    /* Set line number to 0 */
    int j = 0;
    
    /* Loop through the tokenized lines */
    while (tok != NULL) {
        /* Check to see if the line number is matched */
        if (j == line) {
            /* Free the memory allocated previously before
             * returning to caller */
            free(ts);
            
            /* Clear out allocated memory */
            memset(outbuf, 0, size);
            strncpy(outbuf, tok, size);
            
            /* Return TRUE */
            return TRUE;
        }
        
        /* Get the next line from the token structure */
        tok = strtok_r(NULL, AUTHN_CITADEL_LINEDELIM, &ts_state);
        j++;
    }
    
    /* If this code is executing, the line was not found. Return false.
     * Free the memory allocated previously before returning to the caller. */
    free(ts);
    
    return FALSE;
}

/*
 * Low level reader for the Citadel protocol. Reads results into a character
 * buffer, up to size.
 */
static int authn_citadel_llread(BIO *bio, char *result, int size) {
    char rbuf[size];
    time_t ts, tc;
    
    /* Get the current time */
    time(&ts);
    
    /* Set buffer and running buffer to NULL */
    memset(result, 0, size);
    memset(rbuf, 0, size);
    
    /* Loop until break */
    for (;;) {
        /* Clear out the BIO retry flag */
        BIO_clear_retry_flags(bio);
        
        /* Read from the socket */
        int rv = BIO_read(bio, rbuf, size);
        
        /* Get the current time */
        time(&tc);
        
        /* BIO returned characters from the socket */
        if (rv > 0) {
            /* Copy the returned characters into a buffer */
            strncpy(result, rbuf, rv);
            
#ifdef AUTHN_CITADEL_DEBUGMODE
            printf("R: %s", rbuf);
#endif
            
            /* Return with number of characters read from the
             * socket to the calling process. */
            return rv;
        }
        
        /* Determine if there was an error with no retry flag */
        if (rv <= 0 && BIO_should_retry(bio) == FALSE) {
            /* Return BIO error code */
            return rv;
        }
        
        /* Determine if there was a timeout */
        if ((time_t) difftime(tc, ts) > AUTHN_CITADEL_NETTIMEOUT) {
            /* No characters read, and there was a timeout.
             * Return 0 characters read. */
            return 0;
        }
        
    }
    
    /* If this code has been reached, no data was read. Return 0. */
    return 0;
}

/*
 * High level Citadel protocol reader. Gracefully handles conditions such as
 * listing follows conditions.
 */
static int authn_citadel_read(BIO *bio, char *result, int size) {
    char tmpbuf[size];
    char retbuf[size];
    int rv;
    int offset;
    int citret;
    time_t ts, tc;
    
    /* Get current time */
    time(&ts);
    
    /* Clear buffer to NULL */
    memset(tmpbuf, 0, size);
    
    /* Set offset and current Citadel Server return code */
    offset = 0;
    citret = -1;
    
    /* Loop until break */
    for (;;) {
        /* Get the current time */
        time(&tc);
        
        /* If the offset is -1, we have a problem. Stop reading and break! */
        if (offset >= size-1) break;
        
        /* Read from the socket */
        rv = authn_citadel_llread(bio, tmpbuf+offset, size-offset);
        
        /* Determine if characters have been received */
        if (rv > 0) {
            /* Add length of returned characters to offset */
            offset+=rv;
            
            /* Update the timeout to the current time */
            time(&ts);
            
            /* If citadel return code is <0, it has not been updated. Update
             * with the Citadel return code. */
            if (citret < 0) citret = authn_citadel_parse_return_code(tmpbuf);
        }
        
        /* If there was an error (authn_citadel_llread < 0), return to the
         * calling procedure. */
        if (rv < 0) {
            return rv;
        }
        
        /* If the Citadel return code is >= 0, check for the code and determine
         * if further processing is necessary. Note that binary/more data
         * transmissions are unhandled, but are not necessary for this
         * client's implementation. */
        if (citret >= 0) {
            if (citret == AUTHN_CITADEL_STAT_BINARY_FOLLOWS) {
                /* Binary data follows. (TODO: Unhandled) */
                break;
            } else if (citret == AUTHN_CITADEL_STAT_CIT_OK) {
                /* Citadel returned OK. Break from loop. */
                break;
            } else if (citret == AUTHN_CITADEL_STAT_ERROR) {
                /* Citadel returned an error. Break from loop. */
                break;
            } else if (citret == AUTHN_CITADEL_STAT_LISTING_FOLLOWS) {
                /* Citadel has a listing to follow. Read until end of list
                 * has been received. */
                if (authn_citadel_parse_listing_over(tmpbuf) == TRUE) break;
            } else if (citret == AUTHN_CITADEL_STAT_MORE_DATA) {
                /* More data follows. (TODO: Unhandled) */
                break;
            } else if (citret == AUTHN_CITADEL_STAT_SEND_BINARY) {
                /* Citadel wants binary data from client. */
                break;
            } else if (citret == AUTHN_CITADEL_STAT_SEND_LISTING) {
                /* Citadel wants a listing from the client. */
                break;
            } else if (citret == AUTHN_CITADEL_STAT_START_CHAT_MODE) {
                /* Citadel wants to start chat mode. */
                break;
            } else {
                /* Another value was received. Not handled; break. */
                break;
            }
        }
        
        /* Determine if there has been a timeout. If the process times out,
         * break out of the loop. */
        if ((time_t) difftime(tc, ts) > AUTHN_CITADEL_NETTIMEOUT) {
            break;
        }
    }
    
    /* If data has been received, move it into the results buffer. */
    if (offset > 0) {
        memcpy(result, tmpbuf, offset);
    }
    
    /* Return the number of characters read into the results buffer. */
    return offset;
}

/*
 * Write data to the Citadel server
 */
static int authn_citadel_write(BIO *bio, char *out) {
    time_t starttime;
    time_t currenttime;
    
    char outln[strlen(out) + 2];
    
    /* Append a newline to the outline */
    strcpy(outln, out);
    strcat(outln, "\n");
    
    /* Get the current time */
    time(&starttime);
    
    /* Loop until break */
    for (;;) {
        /* Get the current time */
        time(&currenttime);
        
        /* Write to the socket */
        int rv = BIO_write(bio, outln, strlen(outln));
        
#ifdef AUTHN_CITADEL_DEBUGMODE
        printf("W: %s\n", outln);
#endif
        
        /* Flush the socket buffer */
        BIO_flush(bio);		/* Flush BIO buffer */
        
        /* If characters written, return number of characters written
         * to the caller. */
        if (rv > 0) {
            return rv;
        }
        
        /* Check for a timeout condition */
        if ((time_t) difftime(currenttime, starttime) >
                AUTHN_CITADEL_NETTIMEOUT) {
            return 0;
        }
        
        /* Check for an error condition that should not be retried */
        if (rv <= 0 && BIO_should_retry(bio) == FALSE) {
            return rv;
        }
    }
    
    /* If this code has executed, there was an issue. Return 0 characters. */
    return 0;
}

/*
 * Send a command to the citadel server, and read the results into a buffer
 */
static int authn_citadel_send_command(BIO *bio, char *cmd, char *result,
        int varsize) {
    const int _TRANSIZE = varsize - 1;
    int nstat = 0;
    char *ret = NULL;
    char rbuf[_TRANSIZE+1];
    size_t rlen = 0;
    size_t blen = 0;
    size_t rpeek = 0;
    int timeout = 0;
    time_t starttime, currenttime;
    
    /* Ensure BIO is not null. Return error if it is. */
    if (bio == NULL) {
        return -1;
    }
    
    /* Write the command to the citadel server. Return with error if an
     * error code is received. */
    if (authn_citadel_write(bio, cmd) <= 0) {
        return -1;
    }
    
    /* Clear the results buffer to NULL */
    memset(result, 0, varsize);
    
    /* Get the current time */
    time(&starttime);
    
    /* Loop until break */
    for (;;) {
        /* Get the current time */
        time(&currenttime);
        
        /* Read results from the citadel server */
        int rv = authn_citadel_read(bio, rbuf, _TRANSIZE);
        
        /* If the number of characters received is more than 0,
         * copy them to the results buffer and return the
         * number of characters read to the calling process. */
        if (rv > 0) {
            memcpy(result, rbuf, rv);
            return rv;
        }
        
        /* Check for timeout conditions */
        if ((time_t) difftime(currenttime, starttime) >
        AUTHN_CITADEL_NETTIMEOUT) {
            return -1;
        }
    }
    
    /* If this code is executing, there was a problem reading to the buffer.
     * Return 0 characters read */
    return 0;
}

/*
 * Free the BIO references and structures from memory. Pop all chained BIOs
 * before freeing the BIO.
 */
static void authn_citadel_freebio(BIO* bio, int biorefcount) {
    int x;
    
    /* Dereference pushed BIOs */
    for (x = 0; x < biorefcount; x++) {
        /* Pop the BIO */
        bio = BIO_pop(bio);
    }
    
    /* Reset and free the BIO */
    BIO_reset(bio);
    BIO_free_all(bio);
}

/*
 * Concatenate a command and parameters
 */
static void authn_citadel_cat_command(char *result, int reslen, int numparam,
        char *cmd, ...) {
    const int _TRANSIZE = 254;
    int index = numparam;
    va_list vars;
    char ret[_TRANSIZE];
    
    /* Concatenate command */
    strncpy(ret, cmd, _TRANSIZE);
    strncat(ret, " ", _TRANSIZE);
    
    /* Initialize the variable argument list */
    va_start(vars, cmd);
    
    /* Loop through all of the variable arguments */
    while (index--) {
        /* Get the current argument */
        char *s = va_arg(vars, char *);
        
        /* Concatenate the argument onto the Citadel command,
         * delimited by a pipe symbol ("|") */
        strncat(ret, s, _TRANSIZE);
        if (index > 0) strncat(ret, "|", _TRANSIZE);
    }
    
    /* End the variable argument list, free structures */
    va_end(vars);
    
    /* Copy the command into the return buffer */
    strncpy(result, ret, reslen);
}

/*
 * Check the password and username provided by the Apache web server
 */
static authn_status authn_citadel_check_password(request_rec *r,
        const char *user, const char *password) {
    /* Read the module configuration into the conf structure */
    authn_citadel_config_rec *conf = ap_get_module_config(r->per_dir_config,
            &authn_citadel_module);
    
    ap_configfile_t *f; /* Apache configuration file */
    apr_status_t status; /* APR status */
    char hostnameport[254]; /* Host name and port */
    char cmdline[254]; /* Server command line buffer */
    char resbuf[1024]; /* Results buffer */
    char parmbuf[254]; /* Parameter extraction buffer */
    time_t starttime; /* Start time */
    time_t currenttime; /* Current time */
    int lc; /* Result code */
    char realm[254]; /* Citadel realm */
    int brc; /* Bio stack level */
    int rul; /* Required user level */
    int uul; /* User's current level */
    
    /* Reference starts as 0 for BIO pushes */
    brc = 0;
    
    /* Set up OPENSSL Connection */
    BIO* bio = NULL;
    BIO* ssl_bio = NULL;
    
    /* Initialize the SSL library */
    SSL_load_error_strings();
    ERR_load_BIO_strings();
    OpenSSL_add_all_algorithms();
    
    /* Format host:port */
    snprintf(hostnameport, sizeof(hostnameport), "%s:%s",
            conf->citadelserver, conf->port);
    
    /* Initialize SSL pointers */
    SSL_CTX * ctx = SSL_CTX_new(SSLv23_client_method());
    SSL * ssl = SSL_new(ctx);
    SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY);
    
    /* Open the socket connection via OpenSSL libraries */
    bio = BIO_new_connect(hostnameport);
    ssl_bio=BIO_new(BIO_f_ssl());
    BIO_set_ssl(ssl_bio, ssl, BIO_CLOSE);
    
    
    /* Check for success of object creation */
    if (bio == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, status, r,
                "Error allocating citadel server connection object: %s",
                hostnameport);
        return AUTH_GENERAL_ERROR;
    }
    
    /* Set non-blocking I/O */
    BIO_set_nbio(bio, 1);
    
    /* Get the current time */
    time(&starttime);
    
    /* Connect to the Citadel server */
    for (;;) {
        /* Get the current time */
        time(&currenttime);
        
        /* Try to connect */
        int cs = BIO_do_connect(bio);
        
        /* Connection was not made, but should be retried. Note that the
         * loop and retries are necessary as we are using non-blocking
         * I/O for this BIO. */
        if (cs <= 0 && BIO_should_retry(bio) == TRUE) {
            /* Check for timeout conditions */
            if (difftime(starttime, currenttime) > AUTHN_CITADEL_NETTIMEOUT) {
                /* Free the connection */
                authn_citadel_freebio(bio, brc);
                authn_citadel_freebio(ssl_bio, 1);
                
                /* Log the timeout */
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, status, r,
                        "Timeout while trying to open connection: %s",
                        hostnameport);
                
                /* Return a general error. Stop Apache from
                 * authenticating. */
                return AUTH_GENERAL_ERROR;
            } /* Connection was not made, and should not be retried */
        } else if (cs <= 0 && BIO_should_retry(bio) == FALSE) {
            /* Free the connection */
            authn_citadel_freebio(bio, brc);
            authn_citadel_freebio(ssl_bio, 1);
            
            /* Log the error */
            ap_log_rerror(APLOG_MARK,
                    APLOG_NOERRNO|APLOG_ERR,
                    status, r,
                    "Could not open socket connection to citadel server: %s",
                    hostnameport);
            
            /* Return a general error. Stop Apache from
             * authenticating. */
            return AUTH_GENERAL_ERROR;
        } else {
            /* Connection made. Break out of loop. */
            break;
        }
    }
    
    /* Validate Connection. Read connect banner. Error if no banner
     * received. */
    if (authn_citadel_read(bio, resbuf, 1023) <= 0) {
        /* No connect banner read. Log error. */
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, status, r,
                "Citadel Server did not return connect banner: %s",
                user,
                conf->citadelserver);
        
        /* Free the connection */
        authn_citadel_freebio(bio, brc);
        authn_citadel_freebio(ssl_bio, 1);
        
        /* Return general error. Stop Apache from authenticating. */
        return AUTH_GENERAL_ERROR;
    }
    
    /* Check for server OK response on connect. */
    lc = authn_citadel_parse_return_code(resbuf);
    
    /* If there was an error, process it and stop authentication. */
    if (lc != AUTHN_CITADEL_STAT_CIT_OK) {
        /* Log the error. */
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, status, r,
                "Citadel Server did not return OK on connect: %s",
                conf->citadelserver);
        
        /* Free the connection */
        authn_citadel_freebio(bio, brc);
        authn_citadel_freebio(ssl_bio, 1);
        
        /* Return general error. Stop Apache from authenticating. */
        return AUTH_GENERAL_ERROR;
    }
    
    /* Identify this client to the server */
    authn_citadel_cat_command(cmdline, 254, 5, "IDEN",
            AUTHN_CITADEL_IDEN_DEVELOPER,
            AUTHN_CITADEL_IDEN_CLIENTID,
            AUTHN_CITADEL_IDEN_VERSION,
            AUTHN_CITADEL_IDEN_DESCRIPTION,
            "");
    authn_citadel_send_command(bio, cmdline, resbuf, 1023);
    
    /* Get result code and check for OK return status or error. */
    lc = authn_citadel_parse_return_code(resbuf);
    if (lc != AUTHN_CITADEL_STAT_CIT_OK) {
        /* Log the error */
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, status, r,
                "Citadel Server returned failure on IDEN: %s",
                conf->citadelserver);
        
        /* Free the connection */
        authn_citadel_freebio(bio, brc);
        authn_citadel_freebio(ssl_bio, 1);
        
        /* Return general error. Stop Apache from authenticating. */
        return AUTH_GENERAL_ERROR;
    }
    
    /* Perform TLS negotiation, if enabled in the end-user configuration.
     * This is accomplished by appending an SSL BIO to the intial BIO.
     * Once chained together, all communication will be SSL encrypted. */
    if (strcasecmp(conf->usessl, "yes") == 0) {
        /* Request START TLS from Citadel */
        authn_citadel_cat_command(cmdline, 254, 0, "STLS");
        authn_citadel_send_command(bio, cmdline, resbuf, 1023);
        lc = authn_citadel_parse_return_code(resbuf);
        
        /* If a response other than OK was received, there was an
         * error */
        if (lc != AUTHN_CITADEL_STAT_CIT_OK) {
            /* Log the error */
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, status, r,
                    "Citadel Server returned failure on STLS: %s",
                    conf->citadelserver);
            
            /* Free the connection */
            authn_citadel_freebio(bio, brc);
            authn_citadel_freebio(ssl_bio, 1);
            
            /* Return general error. Stop Apache from authenticating. */
            return AUTH_GENERAL_ERROR;
        }
        
        /* Set the SSL read/write BIO */
        SSL_set_bio(ssl, bio, bio);
        
        /* Declare a variable to hold the SSL_connect return status */
        int sslStat;
        
        /* Get the current time */
        time(&starttime);
        
        /* Loop until break */
        for (;;) {
            /* Get the current time */
            time(&currenttime);
            
            /* Try to establish SSL connectivity */
            sslStat = SSL_connect(ssl);
            
            /* If a connection was establish, break out of the
             * loop and continue */
            if (sslStat != -1) {
                break;
            }
            
            /* Check for timeout conditions */
            if (difftime(starttime, currenttime) > AUTHN_CITADEL_NETTIMEOUT) {
                /* Log the timeout */
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, status, r,
                        "Citadel Server timed out on TLS negotiation: %s",
                        conf->citadelserver);
                
                /* Free the connection */
                authn_citadel_freebio(bio, brc);
                authn_citadel_freebio(ssl_bio, 1);
                
                /* Return general error. Stop Apache from authenticating. */
                return AUTH_GENERAL_ERROR;
            }
        }
        
        /* Push SSL_BIO onto BIO
         * Will cause all communications to be encrypted */
        bio = BIO_push(ssl_bio, bio);
        brc++;  /* Increase reference counter for FREE */
        
        /* Handle SSL protocol or initialization errors */
        if (sslStat != 1) {
            /* Log the error */
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, status, r,
                    "TLS negotiation could not take place for Citadel Server: %s",
                    conf->citadelserver);
            
            /* Free the connection */
            authn_citadel_freebio(bio, brc);
            authn_citadel_freebio(ssl_bio, 1);
            
            /* Return general error. Stop Apache from authenticating. */
            return AUTH_GENERAL_ERROR;
        }
        
        /* Verify that communication is encrypted */
        authn_citadel_cat_command(cmdline, 254, 0, "GTLS");
        authn_citadel_send_command(bio, cmdline, resbuf, 1023);
        lc = authn_citadel_parse_return_code(resbuf);
        
        /* If return code is not OK, communication is not encrypted
         * or some other error has occured. */
        if (lc != AUTHN_CITADEL_STAT_CIT_OK) {
            /* Log the error */
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, status, r,
                    "Citadel Server returned failure on GTLS (err %i): %s",
                    lc, conf->citadelserver);
            
            /* Free the connection */
            authn_citadel_freebio(bio, brc);
            authn_citadel_freebio(ssl_bio, 1);
            
            /* Return general error. Stop Apache from authenticating. */
            return AUTH_GENERAL_ERROR;
        }
    }
    
    /* Initiate test command for response (NOOP) to validate that
     * communication is occuring correctly */
    authn_citadel_cat_command(cmdline, 254, 0, "NOOP");
    authn_citadel_send_command(bio, cmdline, resbuf, 1023);
    
    /* Check response code from the NOOP command. If an error occured,
     * cancel the authentication to fail securely */
    lc = authn_citadel_parse_return_code(resbuf);
    if (lc != AUTHN_CITADEL_STAT_CIT_OK) {
        /* Log the error */
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, status, r,
                "NOOP did not return OK for: %s",
                user,
                conf->citadelserver);
        
        /* Free the connection */
        authn_citadel_freebio(bio, brc);
        authn_citadel_freebio(ssl_bio, 1);
        
        
        /* Return general error. Stop Apache from authenticating. */
        return AUTH_GENERAL_ERROR;
    }
    
    /* Get Citadel server information */
    authn_citadel_cat_command(cmdline, 254, 0, "INFO");
    authn_citadel_send_command(bio, cmdline, resbuf, 1023);
    
    /* Check the return response code */
    lc = authn_citadel_parse_return_code(resbuf);
    
    if (lc != AUTHN_CITADEL_STAT_LISTING_FOLLOWS)  /* Listing must follow */ {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, status, r,
                "INFO command did not return list follows for: %s",
                user,
                conf->citadelserver);
        
        /* Free the connection */
        authn_citadel_freebio(bio, brc);
        authn_citadel_freebio(ssl_bio, 1);
        
        /* Return general error. Stop Apache from authenticating. */
        return AUTH_GENERAL_ERROR;
    }
    
    /* Parse the response from Citadel and ensure that END OF LISTING
     * was received. If an invalid response was received, fail securely
     * by terminating the authentication and return a general error. */
    if (authn_citadel_parse_listing_over(resbuf) == FALSE) {
        /* Free the connection */
        authn_citadel_freebio(bio, brc);
        authn_citadel_freebio(ssl_bio, 1);
        
        /* Return general error. Stop Apache from authenticating. */
        return AUTH_GENERAL_ERROR;
    }
    
    /* Pass the username received from Apache to the Citadel server */
    authn_citadel_cat_command(cmdline, 254, 1, "USER", user);
    authn_citadel_send_command(bio, cmdline, resbuf, 1023);
    lc = authn_citadel_parse_return_code(resbuf);
    
    /* Check the USER command for a valid response. If the response was
     * not OK, the user does not exist. Return AUTH_DENIED if user does
     * not exist. */
    if (lc != AUTHN_CITADEL_STAT_MORE_DATA) {
        /* Log the failed authentication attempt. */
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, 0, r,
                "User %s does not exist on Citadel Server: %s",
                user,
                conf->citadelserver);
        
        /* Terminate the session gracefully by sending the QUIT command */
        authn_citadel_cat_command(cmdline, 254, 0, "QUIT");
        authn_citadel_send_command(bio, cmdline, resbuf, 1023);
        
        /* Free the connection */
        authn_citadel_freebio(bio, brc);
        authn_citadel_freebio(ssl_bio, 1);
        
        /* Return AUTH_DENIED to Apache */
        return AUTH_DENIED;
    }
    
    /* Pass the password received from Apache to the Citadel server */
    authn_citadel_cat_command(cmdline, 254, 1, "PASS", password);
    authn_citadel_send_command(bio, cmdline, resbuf, 1023);
    lc = authn_citadel_parse_return_code(resbuf);
    
    /* Check the PASS command for a valid response. If a response other than
     * OK was received, the password was invalid. If the password is
     * invalid, return AUTH_DENIED. */
    if (lc != AUTHN_CITADEL_STAT_CIT_OK) {
        /* Log the failed authentication attempt */
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, status, r,
                "User %s did not submit correct credentials for: %s",
                user,
                conf->citadelserver);
        
        /* Terminate the session gracefully */
        authn_citadel_cat_command(cmdline, 254, 0, "QUIT");
        authn_citadel_send_command(bio, cmdline, resbuf, 1023);
        
        /* Free the connection */
        authn_citadel_freebio(bio, brc);
        authn_citadel_freebio(ssl_bio, 1);
        
        /* Return AUTH_DENIED to the Apache server */
        return AUTH_DENIED;
    }
    
    /* Check the user's access level. The access level is returned as
     * the second parameter on the return line */
    if (authn_citadel_parse_return_parm(parmbuf, 2, resbuf, 254) != -1) {
        rul = 1;
        rul = atoi(conf->minlevel);
        
        uul = -1;
        uul = atoi(parmbuf);
        
        /* If no user level returned, or less then required,
         * deny authentication */
        if (uul < rul) {
            /* Terminate the session gracefully. */
            authn_citadel_cat_command(cmdline, 254, 0, "QUIT");
            authn_citadel_send_command(bio, cmdline, resbuf, 1023);
            
            /* Free the connection */
            authn_citadel_freebio(bio, brc);
            authn_citadel_freebio(ssl_bio, 1);
            
            /* Return AUTH_DENIED to the Apache server */
            return AUTH_DENIED;
        }
    }
    
    
    /* Terminate the session gracefully. */
    authn_citadel_cat_command(cmdline, 254, 0, "QUIT");
    authn_citadel_send_command(bio, cmdline, resbuf, 1023);
    
    /* Free the connection */
    authn_citadel_freebio(bio, brc);
    authn_citadel_freebio(ssl_bio, 1);
    
    /* Authentication was successful. Set the correct status code for the
     * Apache status. */
    status = APR_SUCCESS;
    
    /* Return AUTH_GRANTED to the Apache server. */
    return AUTH_GRANTED;
}

/*
 * State that the authn_citadel_check_password routine is a provider
 * for authentication
 */
static const authn_provider authn_citadel_provider = {
    &authn_citadel_check_password,
    NULL
};

/*
 * Register the citadel authentication provider with Apache
 */
static void register_hooks(apr_pool_t *p) {
    ap_register_provider(p, AUTHN_PROVIDER_GROUP, "citadel", "0",
            &authn_citadel_provider);
    
}

/*
 * Declare the Citadel module data to the Apache loader and server
 */
module AP_MODULE_DECLARE_DATA authn_citadel_module = {
    STANDARD20_MODULE_STUFF,
    create_authn_citadel_dir_config,    /* dir config creater */
    NULL,                               /* dir config merger */
    NULL,                               /* server config creator */
    NULL,                               /* server config merger */
    authn_citadel_cmds,                 /* command apr_table_t */
    register_hooks                      /* register hooks */
};


